
#ifndef __EDITOR_WINDOW_H__
#define __EDITOR_WINDOW_H__

//#include "sidebar.h"
#include "info_bar.h"
#include "map_canvas.h"

class EditorWindow : public wxFrame
{
private:
	//SideBar		*side_bar;
	InfoBar		*info_bar;
	MapCanvas	*map_area;

	wxMenu		*line_context_menu;
	wxMenu		*sector_context_menu;

protected:

public:
	enum
	{
		// File menu
		EWMENU_FILE_WADMANAGER,
		EWMENU_FILE_SAVE,
		EWMENU_FILE_SAVEAS,
		EWMENU_FILE_SAVESTANDALONE,
		EWMENU_FILE_RENAME,
		EWMENU_FILE_CLOSE,
		EWMENU_FILE_EXIT,

		// Line edit menu
		EWMENU_EDIT_LINE,
		EWMENU_EDIT_ALIGNX,
		EWMENU_EDIT_ALIGNY,
		EWMENU_EDIT_CORRECTREFS,
		EWMENU_EDIT_SPLITLINE,
		EWMENU_EDIT_EXTRUDELINE,

		// Sector edit menu
		EWMENU_EDIT_SECTOR,
		EWMENU_EDIT_CREATESECTOR,
		EWMENU_EDIT_MERGESECTORS,
		EWMENU_EDIT_JOINSECTORS,
		EWMENU_EDIT_CREATEDOOR,
		EWMENU_EDIT_CREATESTAIRS,
		EWMENU_EDIT_LIGHTGRADIENT,

		// Edit menu
		EWMENU_EDIT_SCRIPTS,
		EWMENU_EDIT_PREFS,

		// Mode menu
		EWMENU_MODE_VERTS,
		EWMENU_MODE_LINES,
		EWMENU_MODE_SECTORS,
		EWMENU_MODE_THINGS,
		EWMENU_MODE_3D,

		// View menu
		EWMENU_VIEW_CONSOLE,
		EWMENU_VIEW_SETBGTEX,
		EWMENU_VIEW_CLEARBGTEX,

		// Checks menu
		EWMENU_CHECK_REMOVEVERTS,
		EWMENU_CHECK_REMOVELINES,
		EWMENU_CHECK_REMOVESECTORS,
		EWMENU_CHECK_REMOVESIDES,
		EWMENU_CHECK_TEXTURES,
	};

	EditorWindow(const wxChar *title, int xpos, int ypos, int width, int height);
	~EditorWindow();

	void setup_menu();
	void setup_toolbar();

	bool open_map(Wad* wad, string mapname);
	void redraw_map(bool map = false, bool grid = false) { map_area->redraw(map, grid); }
	void update_infobar() { info_bar->update(); }
	void update_statusbar();
	void change_infobar();
	void popup_context_menu() { map_area->popup_menu(); }

	// Window events
	//void key_down(wxKeyEvent &event);
	//void key_up(wxKeyEvent &event);
	void close(wxCloseEvent &event);

	// Menu events
	void file_wadmanager(wxCommandEvent &event);
	void file_close(wxCommandEvent &event);
	void file_exit(wxCommandEvent &event);
	void file_save(wxCommandEvent &event);
	void file_saveas(wxCommandEvent &event);
	void file_savestandalone(wxCommandEvent &event);
	void file_rename(wxCommandEvent &event);

	void edit_scripts(wxCommandEvent &event);
	void edit_prefs(wxCommandEvent &event);

	void edit_alignx(wxCommandEvent &event);
	void edit_correctrefs(wxCommandEvent &event);
	void edit_splitline(wxCommandEvent &event);
	void edit_extrudeline(wxCommandEvent &event);

	void edit_mergesectors(wxCommandEvent &event);
	void edit_joinsectors(wxCommandEvent &event);
	void edit_createdoor(wxCommandEvent &event);
	void edit_createstairs(wxCommandEvent &event);
	void edit_createsector(wxCommandEvent &event);
	void edit_lightgradient(wxCommandEvent &event);

	void mode_vertices(wxCommandEvent &event);
	void mode_lines(wxCommandEvent &event);
	void mode_sectors(wxCommandEvent &event);
	void mode_things(wxCommandEvent &event);
	void mode_3d(wxCommandEvent &event);

	void view_console(wxCommandEvent &event);
	void view_setbgtex(wxCommandEvent &event);
	void view_clearbgtex(wxCommandEvent &event);

	void check_removeverts(wxCommandEvent &event);
	void check_removelines(wxCommandEvent &event);
	void check_removesectors(wxCommandEvent &event);
	void check_removesides(wxCommandEvent &event);
	void check_checktextures(wxCommandEvent &event);

	DECLARE_EVENT_TABLE()
};

#endif
