
#include "main.h"
#include "map_canvas.h"
#include "map.h"
#include "draw.h"
#include "edit.h"
#include "textures.h"
#include "keybind.h"
#include "input.h"
#include "linedraw.h"
#include "undoredo.h"
#include "copypaste.h"
#include "editor_window.h"

bool thing_quickangle = false;
bool items_moving = false;
bool paste_mode = false;

vector<string> pressed_keys;
vector<string> released_keys;

rect_t sel_box(-1, -1, -1, -1);

wxGLContext *gl_context = NULL;

extern int vid_width, vid_height, hilight_item, edit_mode;
extern Map map;
extern point2_t mouse, down_pos;
extern bool allow_tex_load, line_draw;
extern BindList binds;
extern Clipboard clipboard;
extern EditorWindow *editor_window;
extern rgba_t col_line_solid;

EXTERN_CVAR(Bool, line_aa)

MapCanvas::MapCanvas(wxWindow *parent)
:	wxGLCanvas(parent, -1)
{
	SetCurrent();
	glViewport(0, 0, GetSize().x, GetSize().y);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glShadeModel(GL_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glCullFace(GL_NONE);
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_FOG);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glOrtho(0.0f, GetSize().x, GetSize().y, 0.0f, -1.0f, 1.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	update_map();
	update_grid();

	allow_tex_load = true;
	wxSafeYield();
	glEnable(GL_TEXTURE_2D);
	init_textures();

	gl_context = GetContext();
}

MapCanvas::~MapCanvas()
{
}

void MapCanvas::redraw(bool map, bool grid)
{
	SetCurrent();

	if (map)
		update_map();

	if (grid)
		update_grid();

	draw_map();

	SwapBuffers();
}

void MapCanvas::popup_menu()
{
	if (edit_mode == 1)
		PopupMenu(pmenu_line, wxPoint(down_pos.x, down_pos.y));

	if (edit_mode == 2)
		PopupMenu(pmenu_sector, wxPoint(down_pos.x, down_pos.y));
}

BEGIN_EVENT_TABLE(MapCanvas, wxGLCanvas)
	EVT_PAINT(MapCanvas::paint)
	EVT_SIZE(MapCanvas::resize)
	EVT_MOUSE_EVENTS(MapCanvas::mouse_event)
	EVT_MOUSEWHEEL(MapCanvas::mouse_wheel)
	EVT_KEY_DOWN(MapCanvas::key_down)
	EVT_KEY_UP(MapCanvas::key_up)
END_EVENT_TABLE()

void MapCanvas::paint(wxPaintEvent &event)
{
	wxPaintDC dc(this);

	redraw(false, false);

	//event.RequestMore();
}

void MapCanvas::resize(wxSizeEvent& event)
{
	vid_width = GetSize().x;
	vid_height = GetSize().y;

	SetCurrent();

	glViewport(0, 0, GetSize().x, GetSize().y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glOrtho(0.0f, GetSize().x, GetSize().y, 0.0f, -1.0f, 1.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	redraw(true, true);

	OnSize(event);
}

void MapCanvas::mouse_event(wxMouseEvent &event)
{
	if (event.Moving() || event.Dragging())
		mouse_motion(event);

	if (event.ButtonDown())
		mouse_down(event);

	if (event.ButtonUp())
		mouse_up(event);

	if (event.ButtonDClick())
		mouse_down(event);

	if (event.Leaving())
	{
		binds.unset("Mouse1", &released_keys, false, false, false);
		binds.unset("Mouse2", &released_keys, false, false, false);
		binds.unset("Mouse3", &released_keys, false, false, false);
		//keys_edit();
		//released_keys.clear();
	}

	mouse_wheel(event);

	redraw(false, false);
}

void MapCanvas::mouse_motion(wxMouseEvent& event)
{
	bool update_map = false;

	mouse.set(event.GetPosition().x, event.GetPosition().y);

	// Selection box
	if (sel_box.x1() != -1)
	{
		sel_box.br.set(mouse.x, mouse.y);

		if (line_draw)
			line_drawbox();
	}
	else
	{
		if (binds.pressed("edit_selectbox"))
		{
			sel_box.tl.set(down_pos);
			sel_box.br.set(down_pos);
		}
	}

	// Moving items
	if (items_moving)
	{
		move_items();
		update_map = true;
	}
	else
	{
		if (binds.pressed("edit_moveitems") && (selection() || hilight_item != -1))
		{
			items_moving = true;
			add_move_items();
			update_map = true;
		}
	}

	// Quick thing angle
	if (thing_quickangle)
	{
		thing_setquickangle();
		update_map = true;
	}
	else
	{
		if (binds.pressed("thing_quickangle"))
		{
			make_backup(false, false, false, false, true);
			thing_quickangle = true;
			thing_setquickangle();
			update_map = true;
		}
	}

	if (!thing_quickangle && !paste_mode && !line_draw && !items_moving && sel_box.x1() == -1)
	{
		int old_hilight = hilight_item;
		get_hilight_item(mouse.x, mouse.y);

		if (old_hilight != hilight_item)
			editor_window->update_infobar();
	}

	editor_window->update_statusbar();
	if (update_map) redraw(true, false);
}

void MapCanvas::mouse_down(wxMouseEvent& event)
{
	SetFocus();

	bool redraw_map = false;
	bool update_map = false;

	if (line_draw)
	{
		if (event.Button(wxMOUSE_BTN_LEFT))
		{
			line_drawpoint();
			redraw_map = true;
		}

		if (event.Button(wxMOUSE_BTN_RIGHT))
		{
			line_undrawpoint();
			redraw_map = true;
		}

		if (event.Button(wxMOUSE_BTN_MIDDLE))
		{
			line_drawpoint(true);
			redraw_map = true;
		}
	}
	else if (paste_mode)
	{
		if (event.Button(wxMOUSE_BTN_LEFT))
		{
			paste_mode = false;
			clipboard.Paste();
			redraw_map = update_map = true;
		}
	}
	else
	{
		down_pos.set(mouse);
		string key;

		if (event.Button(wxMOUSE_BTN_LEFT))
			binds.set("Mouse1", &pressed_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());

		if (event.Button(wxMOUSE_BTN_RIGHT))
			binds.set("Mouse3", &pressed_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());

		if (event.Button(wxMOUSE_BTN_MIDDLE))
			binds.set("Mouse2", &pressed_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());

		keys_edit();
		pressed_keys.clear();
	}

	if (redraw_map)
		redraw(update_map, false);
}

void MapCanvas::mouse_up(wxMouseEvent& event)
{
	if (paste_mode || line_draw)
		return;

	if (event.Button(wxMOUSE_BTN_LEFT))
		binds.unset("Mouse1", &released_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.Button(wxMOUSE_BTN_RIGHT))
		binds.unset("Mouse3", &released_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());

	if (event.Button(wxMOUSE_BTN_MIDDLE))
		binds.unset("Mouse2", &released_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());

	keys_edit();
	released_keys.clear();
}

void MapCanvas::mouse_wheel(wxMouseEvent& event)
{
	string key = "";

	if (event.GetWheelRotation() > 0)
		key = "MWheel Up";
	else if (event.GetWheelRotation() < 0)
		key = "MWheel Down";
	else
		return;

	// 'Press' wheel 'key'
	down_pos.set(mouse);
	binds.set(key, &pressed_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());
	keys_edit();
	pressed_keys.clear();

	// 'Release' wheel 'key'
	binds.unset(key, &released_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());
	keys_edit();
	released_keys.clear();
}

void MapCanvas::key_down(wxKeyEvent &event)
{
	//wxLogMessage("Key down event");
	released_keys.clear();
	pressed_keys.clear();

	down_pos.set(mouse);
	binds.set(get_key_name(event.GetKeyCode()), &pressed_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());
	keys_edit();
	pressed_keys.clear();

	event.Skip();
}

void MapCanvas::key_up(wxKeyEvent &event)
{
	//wxLogMessage("Key up event");
	released_keys.clear();
	pressed_keys.clear();

	binds.unset(get_key_name(event.GetKeyCode()), &released_keys, event.ShiftDown(), event.ControlDown(), event.AltDown());
	keys_edit();
	released_keys.clear();

	event.Skip();
}
