
#include "main.h"
#include "prefs_dialog.h"

extern BindList binds;
extern vector<string> all_controls;

InputTextCtrl::InputTextCtrl(wxWindow *parent, int id)
//:	wxStaticText(parent, -1, "", wxDefaultPosition, wxSize(112, -1), wxSUNKEN_BORDER|wxWANTS_CHARS|wxALIGN_CENTER)
:	wxTextCtrl(parent, id, "", wxDefaultPosition, wxSize(112, -1), wxTE_CENTRE|wxWANTS_CHARS|wxTE_PROCESS_TAB|wxTE_PROCESS_ENTER)
{
	Enable(false);
}

InputTextCtrl::~InputTextCtrl()
{
}

void InputTextCtrl::update()
{
	SetValue(key.c_str());
	//SetLabel(key.c_str());

	wxString smods = _T("Modifiers: ");

	bool none = true;
	if (mods & KMOD_SHIFT)
	{
		none = false;
		smods += "Shift";
	}

	if (mods & KMOD_CTRL)
	{
		if (none)
			smods += "Ctrl";
		else
			smods += "+Ctrl";

		none = false;
	}

	if (mods & KMOD_ALT)
	{
		if (none)
			smods += "Alt";
		else
			smods += "+Alt";

		none = false;
	}

	if (none)
		smods += "None";

	mod_label->SetLabel(smods);
}

void InputTextCtrl::set_bind(keybind_t *b)
{
	bind = b;
	key = b->key;
	mods = b->mods;
	update();
}

void InputTextCtrl::accept_changes()
{
	bind->key = key;
	bind->mods = mods;
}

void InputTextCtrl::cancel_changes()
{
	key = bind->key;
	mods = bind->mods;
	update();
}

void InputTextCtrl::set_default()
{
	bind->key = bind->defaultkey;
	bind->mods = bind->defaultmods;
	cancel_changes();
}

BEGIN_EVENT_TABLE(InputTextCtrl, wxTextCtrl)
	EVT_KEY_DOWN(InputTextCtrl::key_down)
	EVT_CHAR(InputTextCtrl::key_char)
	EVT_MOUSE_EVENTS(InputTextCtrl::mouse_event)
END_EVENT_TABLE()

void InputTextCtrl::key_down(wxKeyEvent &event)
{
	key = get_key_name(event.GetKeyCode());

	mods = 0;

	if (event.ShiftDown())
		mods |= KMOD_SHIFT;

	if (event.ControlDown())
		mods |= KMOD_CTRL;

	if (event.AltDown())
		mods |= KMOD_ALT;

	update();
}

void InputTextCtrl::key_char(wxKeyEvent &event)
{
	if (event.GetKeyCode() == WXK_TAB)
	{
		key = get_key_name(WXK_TAB);

		mods = 0;

		if (event.ShiftDown())
			mods |= KMOD_SHIFT;

		if (event.ControlDown())
			mods |= KMOD_CTRL;

		if (event.AltDown())
			mods |= KMOD_ALT;
	}

	update();
	event.Skip();
}

void InputTextCtrl::mouse_event(wxMouseEvent &event)
{
	SetFocus();

	// Mouse buttons
	if (event.ButtonDown())
	{
		if (event.Button(wxMOUSE_BTN_LEFT))
			key = "Mouse1";

		if (event.Button(wxMOUSE_BTN_RIGHT))
			key = "Mouse3";

		if (event.Button(wxMOUSE_BTN_MIDDLE))
			key = "Mouse2";

		if (event.ShiftDown())
			mods |= KMOD_SHIFT;

		if (event.ControlDown())
			mods |= KMOD_CTRL;

		if (event.AltDown())
			mods |= KMOD_ALT;

		update();
		return;
	}

	// Mouse wheel
	if (event.GetWheelRotation() > 0)
	{
		key = "MWheel Up";

		if (event.ShiftDown())
			mods |= KMOD_SHIFT;

		if (event.ControlDown())
			mods |= KMOD_CTRL;

		if (event.AltDown())
			mods |= KMOD_ALT;

		update();
		return;
	}
	else if (event.GetWheelRotation() < 0)
	{
		key = "MWheel Down";

		if (event.ShiftDown())
			mods |= KMOD_SHIFT;

		if (event.ControlDown())
			mods |= KMOD_CTRL;

		if (event.AltDown())
			mods |= KMOD_ALT;

		update();
		return;
	}
}


InputPrefs::InputPrefs(wxWindow *parent)
:	wxPanel(parent, -1)
{
	wxBoxSizer *m_hbox = new wxBoxSizer(wxHORIZONTAL);
	SetSizer(m_hbox);

	// Controls frame
	wxStaticBox *frame = new wxStaticBox(this, -1, "Controls");
	wxStaticBoxSizer *box = new wxStaticBoxSizer(frame, wxHORIZONTAL);
	m_hbox->Add(box, 0, wxEXPAND|wxALL, 4);

	list_controls = new wxListBox(this, IP_LIST_CONTROLS);
	box->Add(list_controls, 1, wxEXPAND|wxALL, 4);

	for (int a = 0; a < binds.n_keys(); a++)
		list_controls->Append(binds.get_bind(a)->name.c_str());

	wxBoxSizer *vbox = new wxBoxSizer(wxVERTICAL);
	m_hbox->Add(vbox, 0, wxEXPAND);

	// Binding frame
	frame = new wxStaticBox(this, -1, "Binding");
	box = new wxStaticBoxSizer(frame, wxVERTICAL);
	vbox->Add(box, 0, wxEXPAND|wxALL, 4);

	entry_binding = new InputTextCtrl(this, IP_ENTRY_BINDING);
	box->Add(entry_binding, 0, wxEXPAND|wxBOTTOM, 4);

	label_modifiers = new wxStaticText(this, IP_LABEL_MODS, "Modifiers:");
	entry_binding->set_modlabel(label_modifiers);
	box->Add(label_modifiers, 0, wxEXPAND|wxBOTTOM, 4);

	btn_set_accept = new wxButton(this, IP_BTN_SETACCEPT, "Set");
	box->Add(btn_set_accept, 0, wxEXPAND|wxBOTTOM, 4);

	btn_default_cancel = new wxButton(this, IP_BTN_DEFAULTCANCEL, "Default");
	box->Add(btn_default_cancel, 0, wxEXPAND);

	// Config frame
	frame = new wxStaticBox(this, -1, "Configuration");
	box = new wxStaticBoxSizer(frame, wxVERTICAL);
	vbox->Add(box, 0, wxEXPAND|wxALL, 4);

	btn_defaults = new wxButton(this, IP_BTN_DEFAULTS, "Restore Defaults");
	box->Add(btn_defaults, 0, wxEXPAND|wxBOTTOM, 4);

	btn_loadconfig = new wxButton(this, IP_BTN_LOADCONFIG, "Load Config");
	box->Add(btn_loadconfig, 0, wxEXPAND|wxBOTTOM, 4);

	btn_saveconfig = new wxButton(this, IP_BTN_SAVECONFIG, "Save Config");
	box->Add(btn_saveconfig, 0, wxEXPAND|wxBOTTOM, 4);
}

InputPrefs::~InputPrefs()
{
}

BEGIN_EVENT_TABLE(InputPrefs, wxPanel)
	EVT_LISTBOX(IP_LIST_CONTROLS, InputPrefs::list_controls_changed)
	EVT_BUTTON(IP_BTN_SETACCEPT, InputPrefs::btn_setaccept_clicked)
	EVT_BUTTON(IP_BTN_DEFAULTCANCEL, InputPrefs::btn_defaultcancel_clicked)
	EVT_BUTTON(IP_BTN_DEFAULTS, InputPrefs::btn_defaults_clicked)
	EVT_BUTTON(IP_BTN_SAVECONFIG, InputPrefs::btn_saveconfig_clicked)
	EVT_BUTTON(IP_BTN_LOADCONFIG, InputPrefs::btn_loadconfig_clicked)
	EVT_BUTTON(IP_BTN_ADDCONTROL, InputPrefs::btn_addcontrol_clicked)
	EVT_TEXT_ENTER(IP_ENTRY_BINDING, InputPrefs::entry_binding_enter)
END_EVENT_TABLE()

void InputPrefs::entry_binding_enter(wxCommandEvent &event)
{
	entry_binding->set_key("Return");
	entry_binding->update();
	event.Skip();
}

void InputPrefs::list_controls_changed(wxCommandEvent &event)
{
	entry_binding->set_bind(binds.get_bind(list_controls->GetSelection()));
}

void InputPrefs::btn_setaccept_clicked(wxCommandEvent &event)
{
	if (list_controls->GetSelection() == wxNOT_FOUND)
		return;

	if (btn_set_accept->GetLabel() == "Set")
	{
		btn_set_accept->SetLabel("Accept");
		btn_default_cancel->SetLabel("Cancel");
		entry_binding->Enable(true);
		entry_binding->SetFocus();
	}
	else
	{
		btn_set_accept->SetLabel("Set");
		btn_default_cancel->SetLabel("Default");
		entry_binding->accept_changes();
		entry_binding->Enable(false);
	}
}

void InputPrefs::btn_defaultcancel_clicked(wxCommandEvent &event)
{
	if (list_controls->GetSelection() == wxNOT_FOUND)
		return;

	if (btn_default_cancel->GetLabel() == "Default")
	{
		entry_binding->set_default();
	}
	else
	{
		btn_default_cancel->SetLabel("Default");
		btn_set_accept->SetLabel("Set");
		entry_binding->cancel_changes();
		entry_binding->Enable(false);
	}
}

void InputPrefs::btn_defaults_clicked(wxCommandEvent &event)
{
	binds.set_defaults();
	entry_binding->update();
}

void InputPrefs::btn_loadconfig_clicked(wxCommandEvent &event)
{
	string filename = wx_to_str(wxFileSelector("Open Key Configuration", "./config/keys", "", "*.cfg",
									"Configuration Files|*.cfg", wxOPEN|wxFILE_MUST_EXIST));

	if (filename != "")
	{
		Tokenizer tz;
		tz.open_file(filename, 0, 0);
		binds.load(&tz);
	}
}

void InputPrefs::btn_saveconfig_clicked(wxCommandEvent &event)
{
	string filename = wx_to_str(wxFileSelector("Save Key Configuration", "./config/keys", "", "*.cfg",
									"Configuration Files|*.cfg", wxSAVE|wxOVERWRITE_PROMPT));

	if (filename != "")
	{
		FILE *fp = fopen(filename.c_str(), "wt");
		binds.save(fp);
		fclose(fp);
	}
}

void InputPrefs::btn_addcontrol_clicked(wxCommandEvent &event)
{
}
