
#include "main.h"
#include "prefs_dialog.h"
#include "editor_window.h"

extern EditorWindow *editor_window;

EXTERN_CVAR(Float, line_size)
EXTERN_CVAR(Bool, line_aa)
EXTERN_CVAR(Int, crosshair_2d)
EXTERN_CVAR(Bool, thing_sprites)
EXTERN_CVAR(Bool, thing_force_angle)
EXTERN_CVAR(Bool, grid_dashed)
EXTERN_CVAR(Bool, grid_64grid)
EXTERN_CVAR(Bool, edit_snap_grid)
EXTERN_CVAR(Bool, allow_np2_tex)
EXTERN_CVAR(Int, tex_filter)

VisualPrefs::VisualPrefs(wxWindow *parent)
:	wxPanel(parent, -1)
{
	wxBoxSizer *m_vbox = new wxBoxSizer(wxVERTICAL);
	SetSizer(m_vbox);

	// Editor frame
	wxStaticBox *frame = new wxStaticBox(this, -1, "Editor");
	wxStaticBoxSizer *box = new wxStaticBoxSizer(frame, wxHORIZONTAL);
	m_vbox->Add(box, 0, wxEXPAND|wxALL, 4);

	wxBoxSizer *vbox = new wxBoxSizer(wxVERTICAL);
	box->Add(vbox, 1, wxEXPAND);

	cb_grid_dash = new wxCheckBox(this, VP_GRID_DASH, "Dashed grid");
	if (grid_dashed) cb_grid_dash->SetValue(true);
	vbox->Add(cb_grid_dash, 0, wxBOTTOM, 2);

	cb_64_grid = new wxCheckBox(this, VP_64_GRID, "Show 64 grid");
	if (grid_64grid) cb_64_grid->SetValue(true);
	vbox->Add(cb_64_grid, 0, wxBOTTOM, 2);

	cb_snap_to_grid = new wxCheckBox(this, VP_SNAP_GRID, "Snap to grid");
	if (edit_snap_grid) cb_snap_to_grid->SetValue(true);
	vbox->Add(cb_snap_to_grid, 0, wxBOTTOM, 2);

	cb_thing_sprites = new wxCheckBox(this, VP_THING_SPRITES, "Show things as sprites");
	if (thing_sprites) cb_thing_sprites->SetValue(true);
	vbox->Add(cb_thing_sprites, 0, wxBOTTOM, 2);

	cb_thing_angles = new wxCheckBox(this, VP_THING_ANGLES, "Force thing angles");
	if (thing_force_angle) cb_thing_angles->SetValue(true);
	vbox->Add(cb_thing_angles, 0, wxBOTTOM, 2);

	wxBoxSizer *hbox = new wxBoxSizer(wxHORIZONTAL);
	vbox->Add(hbox, 0, wxBOTTOM|wxEXPAND, 2);

	hbox->Add(new wxStaticText(this, -1, "Crosshair:"), 0, wxALIGN_CENTER_VERTICAL|wxRIGHT, 4);

	wxString xhair[] = {
		"None",
		"Full",
		"Small",
	};

	ch_2d_xhair = new wxChoice(this, VP_XHAIR_2D, wxDefaultPosition, wxDefaultSize, 3, xhair);
	ch_2d_xhair->SetSelection(crosshair_2d);
	hbox->Add(ch_2d_xhair, 0, wxEXPAND);

	// Lines frame
	frame = new wxStaticBox(this, -1, "Lines");
	box = new wxStaticBoxSizer(frame, wxHORIZONTAL);
	m_vbox->Add(box, 0, wxEXPAND|wxALL, 4);

	box->Add(new wxStaticText(this, -1, "Width:"), 0, wxRIGHT|wxALIGN_CENTER_VERTICAL, 4);
	sl_line_size = new wxSlider(this, VP_LINE_SIZE, line_size * 10, 10, 20, wxDefaultPosition, wxDefaultSize, wxSL_HORIZONTAL|wxSL_AUTOTICKS);
	box->Add(sl_line_size, 1, wxEXPAND);

	cb_line_aa = new wxCheckBox(this, VP_LINE_AA, "Antialias");
	if (line_aa) cb_line_aa->SetValue(true);
	box->Add(cb_line_aa, 0, wxLEFT|wxALIGN_CENTER_VERTICAL, 4);

	// Textures frame
	frame = new wxStaticBox(this, -1, "Textures");
	box = new wxStaticBoxSizer(frame, wxVERTICAL);
	m_vbox->Add(box, 0, wxEXPAND|wxALL, 4);

	cb_tex_np2 = new wxCheckBox(this, VP_TEX_NP2, "Allow non-power-of-two textures");
	if (allow_np2_tex) cb_tex_np2->SetValue(true);
	box->Add(cb_tex_np2, 0, wxEXPAND|wxBOTTOM, 4);

	hbox = new wxBoxSizer(wxHORIZONTAL);
	box->Add(hbox, 0, wxEXPAND);

	wxString filters[] = {
		"None",
		"Bilinear",
		"Mipmapped",
	};

	ch_tex_filter = new wxChoice(this, VP_TEX_FILTER, wxDefaultPosition, wxDefaultSize, 3, filters);
	ch_tex_filter->SetSelection(tex_filter - 1);
	hbox->Add(new wxStaticText(this, -1, "Filter:"), 0, wxRIGHT|wxALIGN_CENTER_VERTICAL, 4);
	hbox->Add(ch_tex_filter, 1, wxEXPAND);
}

VisualPrefs::~VisualPrefs()
{
}

BEGIN_EVENT_TABLE(VisualPrefs, wxPanel)
	EVT_CHOICE(VP_TEX_FILTER, VisualPrefs::filter_changed)
	EVT_CHOICE(VP_XHAIR_2D, VisualPrefs::xhair_2d_changed)
	EVT_COMMAND_SCROLL(VP_LINE_SIZE, VisualPrefs::line_size_changed)
	EVT_CHECKBOX(VP_GRID_DASH, VisualPrefs::grid_dash_toggled)
	EVT_CHECKBOX(VP_64_GRID, VisualPrefs::grid_64_toggled)
	EVT_CHECKBOX(VP_SNAP_GRID, VisualPrefs::grid_snap_toggled)
	EVT_CHECKBOX(VP_THING_SPRITES, VisualPrefs::thing_sprites_toggled)
	EVT_CHECKBOX(VP_THING_ANGLES, VisualPrefs::thing_angles_toggled)
	EVT_CHECKBOX(VP_LINE_AA, VisualPrefs::line_aa_toggled)
	EVT_CHECKBOX(VP_TEX_NP2, VisualPrefs::tex_np2_toggled)
END_EVENT_TABLE()

void VisualPrefs::grid_dash_toggled(wxCommandEvent &event)
{
	grid_dashed = false;
	if (cb_grid_dash->GetValue()) grid_dashed = true;
	editor_window->redraw_map(false, true);
}

void VisualPrefs::grid_64_toggled(wxCommandEvent &event)
{
	grid_64grid = false;
	if (cb_64_grid->GetValue()) grid_64grid = true;
	editor_window->redraw_map(false, true);
}

void VisualPrefs::grid_snap_toggled(wxCommandEvent &event)
{
	edit_snap_grid = false;
	if (cb_snap_to_grid->GetValue()) edit_snap_grid = true;
}

void VisualPrefs::thing_sprites_toggled(wxCommandEvent &event)
{
	thing_sprites = false;
	if (cb_thing_sprites->GetValue()) thing_sprites = true;
	editor_window->redraw_map(true, false);
}

void VisualPrefs::thing_angles_toggled(wxCommandEvent &event)
{
	thing_force_angle = false;
	if (cb_thing_angles->GetValue()) thing_force_angle = true;
	editor_window->redraw_map(true, false);
}

void VisualPrefs::line_aa_toggled(wxCommandEvent &event)
{
	line_aa = false;
	if (cb_line_aa->GetValue()) line_aa = true;
	editor_window->redraw_map(true, false);
}

void VisualPrefs::tex_np2_toggled(wxCommandEvent &event)
{
	allow_np2_tex = false;
	if (cb_tex_np2->GetValue()) allow_np2_tex = true;
}

void VisualPrefs::filter_changed(wxCommandEvent &event)
{
	tex_filter = ch_tex_filter->GetSelection() + 1;
	purge_textures();
}

void VisualPrefs::xhair_2d_changed(wxCommandEvent &event)
{
	crosshair_2d = ch_2d_xhair->GetSelection();
}

void VisualPrefs::line_size_changed(wxScrollEvent &event)
{
	float val = sl_line_size->GetValue() / 10.0f;
	line_size = val;
	editor_window->redraw_map(true, false);
}
