
#ifndef __TEXTURES_H__
#define __TEXTURES_H__

#define TEX_STRETCH	0x01
#define TEX_NOALPHA	0x02

class Texture
{
protected:
	bool		id_generated;
	GLuint		gl_id;

public:
	BYTE		filter;
	BYTE		flags;
	string		name;
	int			width, height;		// Texture dimansions (in game units)
	int			rwidth, rheight;	// Texture dimensions (in pixels)
	float		t_x, t_y;

	Texture();
	~Texture();

	bool load_file(string filename);
	void gen_from_data(int width, int height, BYTE* data, int swidth = -1, int sheight = -1);
	GLuint get_gl_id();
	void purge();

	virtual bool gen_gl_tex();
};

class SimpleTexture : public Texture
{
private:
	Wad		*data_wad;
	string	lump_name;

public:
	SimpleTexture();
	~SimpleTexture();

	void set_data(Wad* wad, string lump);

	bool gen_gl_tex();
};

struct patch_header_t
{
	short	width;
	short	height;
	short	left;
	short	top;
};

struct col_header_t
{
	BYTE	row_offset;
	BYTE	n_pixels;
};

struct patch_info_t
{
	short	xoff;
	short	yoff;
	Wad*	wad;
	string	patch;
};

class DoomTexture : public Texture
{
private:
	vector<patch_info_t>	patches;

public:
	DoomTexture();
	~DoomTexture();

	void add_patch(short xoff, short yoff, Wad* wad, string patch)
	{
		patch_info_t p;
		p.xoff = xoff;
		p.yoff = yoff;
		p.wad = wad;
		p.patch = patch;
		patches.push_back(p);
	}

	bool gen_gl_tex();
};

class DoomFlat : public Texture
{
private:
	Wad*	data_wad;
	int		data_lump;

public:
	DoomFlat();
	~DoomFlat();

	void set_data(Wad* wad, int index)
	{
		data_wad = wad;
		data_lump = index;
	}

	bool gen_gl_tex();
};

class FileTexture : public Texture
{
private:
	string	filename;

public:
	FileTexture(string filename) { this->filename = filename; }
	~FileTexture();

	bool gen_gl_tex();
};

/*
class Texture
{
private:
	BYTE		bpp;
	BYTE		*data;
	bool		has_alpha;
	bool		gl_tex_generated;
	int			gl_filter;
	GLuint		gl_id;
	WORD		rwidth;
	WORD		rheight;
	bool		sprite;

public:
	WORD		width;
	WORD		height;
	string		name;
	float		t_x, t_y;

	Texture();
	~Texture();

	void setup(string name, BYTE bpp, int width, int height, bool has_alpha = true, bool sprite = false);
	bool load_file(string name, string filename, int filter = 2);

	void add_pixel(int x, int y, BYTE colour)
	{
		if (x < 0 || x >= rwidth)
			return;

		if (y < 0 || y >= rheight)
			return;

		DWORD p = (y * (rwidth)) + x;

		data[p] = colour;
	}

	void add_pixel(int x, int y, rgba_t colour)
	{
		if (x < 0 || x >= rwidth)
			return;

		if (y < 0 || y >= rheight)
			return;

		DWORD p = ((y * (rwidth)) + x) * 4;

		data[p++] = colour.r;
		data[p++] = colour.g;
		data[p++] = colour.b;
		data[p++] = colour.a;
	}

	GLuint get_gl_id();
};

*/

#define TEXTURES_WALLS		1
#define	TEXTURES_FLATS		2
#define TEXTURES_SPRITES	3
#define TEXTURES_EDITOR		4

Texture* get_texture(string name, int type = 0, bool force = false);
void load_textures();
void init_textures();
void load_flats();
void load_sprites();
void load_tx_textures();
void load_hirestex_textures();
void purge_textures();

#endif
